import numpy as np
import tensorflow as tf

from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.decomposition import PCA

import quple
from quple.data_encoding.first_order_expansion import FirstOrderExpansion
from quple.data_encoding.second_order_expansion import SecondOrderExpansion
from quple.data_encoding.encoding_maps import self_product, cosine_product, modified_cosine_product, distance_measure, arithmetic_mean, second_moment
from quple.data_encoding.encoding_maps import sine_cosine_alternate_product, exponential_squared_sum, exponential_cube_sum
from quple.components.data_preparation import prepare_train_val_test
from quple.classifiers.variational_quantum_classifier import VQC
from quple.trial_wavefunction.efficient_su2 import EfficientSU2
from quple.trial_wavefunction.real_amplitudes import RealAmplitudes
from quple.trial_wavefunction.excitation_preserving import ExcitationPreserving
from quple.classifiers.vqc_logger import VQCLogger

# prepare training data
data = np.load('data/hmumu_twojet_0719.npy')
x = data[:,:-1]
y = data[:,-1]
n_qubit = 5
n_event = 1000
trials = 10
batch_size = 64
epochs = 100
preprocessors = [PCA(n_components=n_qubit), StandardScaler(), MinMaxScaler((-1,1))]

first_order_expansion = FirstOrderExpansion(feature_dimension=n_qubit, copies=2)
second_order_expansion_0 = SecondOrderExpansion(feature_dimension=n_qubit, copies=2, entangle_strategy='linear', encoding_map=self_product)
second_order_expansion_1 = SecondOrderExpansion(feature_dimension=n_qubit, copies=2, entangle_strategy='linear', encoding_map=cosine_product)
second_order_expansion_2 = SecondOrderExpansion(feature_dimension=n_qubit, copies=2, entangle_strategy='linear', encoding_map=distance_measure)
second_order_expansion_3 = SecondOrderExpansion(feature_dimension=n_qubit, copies=2, entangle_strategy='linear', encoding_map=arithmetic_mean)
second_order_expansion_4 = SecondOrderExpansion(feature_dimension=n_qubit, copies=2, entangle_strategy='linear', encoding_map=second_moment)
excitation_preserving = ExcitationPreserving(n_qubit=n_qubit, copies=2)
efficient_su2 = EfficientSU2(n_qubit=n_qubit, copies=2)
real_amplitudes = RealAmplitudes(n_qubit=n_qubit, copies=2)

vqc_sets = [VQC(first_order_expansion, real_amplitudes, metrics=['binary_accuracy','AUC']),
			VQC(first_order_expansion, efficient_su2, metrics=['binary_accuracy','AUC']),
			VQC(first_order_expansion, excitation_preserving, metrics=['binary_accuracy','AUC']),
			VQC(second_order_expansion_0, efficient_su2, metrics=['binary_accuracy','AUC']),
			VQC(second_order_expansion_1, efficient_su2, metrics=['binary_accuracy','AUC']),
			VQC(second_order_expansion_2, efficient_su2, metrics=['binary_accuracy','AUC']),
			VQC(second_order_expansion_3, efficient_su2, metrics=['binary_accuracy','AUC']),
			VQC(second_order_expansion_4, efficient_su2, metrics=['binary_accuracy','AUC'])]

from pdb import set_trace
set_trace()
for vqc in vqc_sets:
	logger = VQCLogger()
	x_train, x_val, x_test, y_train, y_val, y_test = prepare_train_val_test(x, y, train_size=n_event, val_size=n_event, test_size=n_event, preprocessors=preprocessors)
	vqc.run(x_train, y_train, x_val, y_val, x_test, y_test, batch_size=batch_size, epochs=epochs, callbacks=[logger])


import cirq
import numpy as np
import sympy as sp
from itertools import combinations
from sympy import default_sort_key

import quple

#reference: https://github.com/tensorflow/quantum/blob/v0.3.0/tensorflow_quantum/python/util.py
def symbols_in_op(op):
    """Returns the set of symbols in a parameterized gate."""
    if isinstance(op, cirq.EigenGate):
        return op.exponent.free_symbols

    if isinstance(op, cirq.FSimGate):
        ret = set()
        if isinstance(op.theta, sp.Basic):
            ret |= op.theta.free_symbols
        if isinstance(op.phi, sp.Basic):
            ret |= op.phi.free_symbols
        return ret

    if isinstance(op, cirq.PhasedXPowGate):
        ret = set()
        if isinstance(op.exponent, sp.Basic):
            ret |= op.exponent.free_symbols
        if isinstance(op.phase_exponent, sp.Basic):
            ret |= op.phase_exponent.free_symbols
        return ret

    raise ValueError("Attempted to scan for symbols in circuit with unsupported"
                     " ops inside. Expected op found in tfq.get_supported_gates"
                     " but found: ".format(str(op)))
    
def symbols_in_expr_map(expr_map, to_str=False):
    all_symbols = set()
    for expr in expr_map:
        if isinstance(expr, sp.Basic):
            all_symbols |= expr.free_symbols
    if to_str:
        return sorted([str(x) for x in all_symbols])
    return sorted(list(all_symbols), key=default_sort_key)

def get_circuit_unflattened_symbols(circuit:'quple.QuantumCircuit', to_str=True):
    if isinstance(circuit, quple.QuantumCircuit):
        expr_map = circuit.expr_map
        if expr_map is not None:
            symbols = quple.symbols_in_expr_map(expr_map, to_str=to_str)
    else:
        symbols = quple.get_circuit_symbols(circuit)
    return symbols


#reference: https://github.com/tensorflow/quantum/blob/v0.3.0/tensorflow_quantum/python/util.py
def get_circuit_symbols(circuit, to_str=True):
    all_symbols = set()
    for moment in circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                all_symbols |= symbols_in_op(op.gate)
    if to_str:
        return sorted([str(x) for x in all_symbols])
    return sorted(list(all_symbols), key=default_sort_key)

def get_circuit_qubits(circuit):
    all_qubits = set()
    for moment in circuit:
        for op in moment:
            all_qubits |= set(op._qubits)
    return sorted(list(all_qubits))

def get_circuit_symbols_in_order(circuit):
    all_symbols = set()
    symbols_in_order = []
    for moment in circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                new_symbols = symbols_in_op(op.gate)
                symbols_in_order += (new_symbols - all_symbols)
                all_symbols |= symbols_in_op(op.gate)
    return symbols_in_order

def sample_final_states(circuit, samples=1, data=None):
    import tensorflow_quantum as tfq
    layer = tfq.layers.State()
    symbols = get_circuit_symbols(circuit)
    if symbols:
        if not data:
            data = np.random.rand(samples, len(symbols))*2*np.pi
        tensor_final_states = layer(circuit, symbol_names=symbols, symbol_values=data)
    else:
        tensor_final_states = layer([circuit]*samples)
    final_states = [tensor_state.numpy() for tensor_state in tensor_final_states]
    return final_states

def sample_density_matrices(circuit, samples=1, data=None):
    final_states = sample_final_states(circuit, samples, data=data)
    density_matrices = [cirq.density_matrix_from_state_vector(fs) for fs in final_states] 
    return density_matrices


def sample_fidelities(circuit, samples=1, data=None):
    sample_states_1 = sample_final_states(circuit, samples, data=data)
    sample_states_2 = sample_final_states(circuit, samples, data=data)
    fidelities = []
    for s1, s2 in zip(sample_states_1, sample_states_2):
        fidelities.append(cirq.fidelity(s1, s2))
    return fidelities

def circuit_fidelity_pdf(circuit, samples=3000, bins=100, data=None):
    data = np.array(sample_fidelities(circuit, samples, data=data))
    pdf = np.histogram(data, bins=bins, range=(0,1), density=True)
    return pdf

def Haar_pdf(samples=3000, bins=100):
    data = np.linspace(0., 1., samples)
    pdf = np.histogram(data, bins=bins, range=(0.,1.), density=True)
    return pdf

def circuit_fidelity_plot(circuit, samples=3000, bins=100, data=None, KL=True):      
    import matplotlib.pyplot as plt
    data_pqc = np.array(sample_fidelities(circuit, samples, data=data))
    daa_Haar = np.linspace(0., 1., samples)
    plt.hist(daa_Haar, bins=bins, range=(0.,1.), alpha=0.5, density=True, label='Haar')
    plt.hist(data_pqc, bins=bins, range=(0.,1.), alpha=0.5, density=True, label='PQC')
    plt.legend()
    plt.xlabel('Fidelity')
    plt.ylabel('Probability')
    if KL:
        import scipy as sp
        pdf_pqc = np.histogram(data_pqc, bins=bins, range=(0.,1.), density=True)
        pdf_Haar = np.histogram(daa_Haar, bins=bins, range=(0.,1.), density=True)
        Kullback_Leibler_divergence = sp.stats.entropy(pdf_pqc[0], pdf_Haar[0])
        plt.title('$D_{KL}=$'+str(Kullback_Leibler_divergence))
    return plt

def circuit_expressibility_measure(circuit, samples = 3000, bins=100, data=None, relative=False):
    import scipy as sp
    pdf_pqc = circuit_fidelity_pdf(circuit, samples, bins, data=data)
    pdf_Haar = Haar_pdf(samples, bins)
    Kullback_Leibler_divergence = sp.stats.entropy(pdf_pqc[0], pdf_Haar[0])
    expressibility = Kullback_Leibler_divergence
    if relative:
        expressibility_idle_circuit = (2**circuit.n_qubit-1)*(np.log(bins))
        expressibility = -np.log(expressibility/expressibility_idle_circuit)
    return expressibility

def Meyer_Wallach_measure(state):
    state = np.array(state)
    size = state.shape[0]
    n = int(np.log2(size))
    Q = 0.
    
    def linear_mapping(b:int, j:int) -> np.ndarray:
        keep_indices = [i for i in range(size) if b == ((i & (1 << j))!=0)]
        return state[keep_indices]
    
    def distance(u:np.ndarray, v:np.ndarray) -> float:
        return np.sum(np.abs(np.outer(u,v) - np.outer(v,u))**2)/2
    
    for k in range(n):
        iota_0k = linear_mapping(0, k)
        iota_1k = linear_mapping(1, k)
        Q += distance(iota_0k, iota_1k)
    Q = (4/n)*Q
    return Q
    
        
def circuit_entangling_measure(circuit, samples=200, data=None):
    final_states = sample_final_states(circuit, samples, data=data)
    mw_measures = [Meyer_Wallach_measure(fs) for fs in final_states]
    return np.mean(mw_measures)


def circuit_von_neumann_entropy(circuit, samples=200, data=None):
    density_matrices = sample_density_matrices(circuit, samples, data=data)
    von_neumann_entropy = [cirq.von_neumann_entropy(dm) for dm in density_matrices]
    return np.mean(von_neumann_entropy)
    
    
def gradient_variance_test(circuits, op, symbol=None):
    import tensorflow_quantum as tfq
    import tensorflow as tf
    """Compute the variance of a batch of expectations w.r.t. op on each circuit that 
    contains `symbol`."""
    resolved_circuits = []
    # Resolve irrelevant symbols:
    for circuit in circuits: 
        symbols = get_circuit_symbols(circuit)
        if not len(symbols) > 0:
            raise ValueError('No symbols found in circuit')
        if not symbol:
            symbol = symbols[0]
            symbols = symbols[1:]
        else:
            symbols.remove(symbol)
        resolver = cirq.ParamResolver({s:np.random.uniform() * 2.0 * np.pi for s in symbols})
        resolved_circuits.append(cirq.protocols.resolve_parameters(circuit, resolver))

    # Setup a simple layer to batch compute the expectation gradients.
    expectation = tfq.layers.Expectation()

    # Prep the inputs as tensors
    circuit_tensor = tfq.convert_to_tensor(resolved_circuits)
    n_circuits = len(resolved_circuits)
    values_tensor = tf.convert_to_tensor(
        np.random.uniform(0, 2 * np.pi, (n_circuits, 1)).astype(np.float32))

    # Use TensorFlow GradientTape to track gradients.
    with tf.GradientTape() as g:
        g.watch(values_tensor)
        forward = expectation(circuit_tensor,
                              operators=op,
                              symbol_names=[symbol],
                              symbol_values=values_tensor)

    # Return variance of gradients across all circuits.
    grads = g.gradient(forward, values_tensor)
    grad_var = tf.math.reduce_std(grads, axis=0)
    return grad_var.numpy()[0]




from abc import ABC, abstractmethod
from inspect import signature
from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)
import sympy as sp
import numpy as np

import cirq
from cirq.ops.gate_features import SingleQubitGate, TwoQubitGate

from quple.circuit.quantum_circuit import QuantumCircuit
from quple.circuit.templates.template_circuit_block import TemplateCircuitBlock
from quple.components.interaction_graphs import interaction_graph
from quple.utils.utils import merge_pqc
from pdb import set_trace
#D. Zhu, Training of quantum circuits on a hybrid quantum computer


#Devices generally operate in the Z basis, so that rotations around the Z axis will become book-keeping measures rather than physical operations on the device.

class ParameterisedCircuit(QuantumCircuit, ABC):
    def __init__(self, n_qubit:int, copies:int=1,
                rotation_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                parameter_symbol:str='θ',
                final_rotation_layer:bool=False,
                auto_assemble:bool=True,
                flatten_circuit:bool=True,
                name:str=None):
        
        super().__init__(n_qubit, name=name)
        self._copies = copies
        self.rotation_blocks = rotation_blocks
        self.entanglement_blocks = entanglement_blocks
        self._entangle_strategy = entangle_strategy or 'full' 
        self._parameter_symbol = parameter_symbol
        self._final_rotation_layer = final_rotation_layer
        self._parameters = np.array([], dtype=object)
        self._readout_qubit = None
        self._flatten_circuit = flatten_circuit
        if auto_assemble:
            self.assemble()
    
    @property
    def readout_qubit(self):
        return self._readout_qubit
            
    @property
    def num_param(self):
        return len(self._parameters)

    def new_param(self, size:int=1) -> sp.Symbol:
        start = self.num_param
        params = np.array([sp.Symbol('%s_%s' % (self.parameter_symbol, i)) \
                          for i in range(start, start+size)])
        self._parameters = np.append(self._parameters, params)
        if len(params) == 1:
            return params[0]
        return params
        
    @property
    def parameters(self):
        return self._parameters
    
    @property
    def flatten_circuit(self):
        return self._flatten_circuit
    
    
    @property
    def copies(self) -> int:
        return self._copies
    
    @property
    def parameter_symbol(self) -> str:
        return self._parameter_symbol
    
    @property
    def entangle_strategy(self):
        return self._entangle_strategy
    
    @property
    def final_rotation_layer(self):
        return self._final_rotation_layer

    @property
    def rotation_blocks(self):
        '''A collection of single qubit gate operations or 
        circuit blocks consisting of multiple single qubit operations
        '''
        return self._rotation_blocks
    
    @rotation_blocks.setter
    def rotation_blocks(self, val):
        val = ParameterisedCircuit._parse_blocks(val)
        ParameterisedCircuit._validate_blocks(val, SingleQubitGate, allow_template=False)
        self._rotation_blocks = val
    
    @property
    def entanglement_blocks(self):
        '''A collection of two qubit gate operations or 
        circuit blocks consisting of multiple single/two qubit operations
        '''
        return self._entanglement_blocks
    
    @entanglement_blocks.setter
    def entanglement_blocks(self, val):
        val = ParameterisedCircuit._parse_blocks(val)
        ParameterisedCircuit._validate_blocks(val, TwoQubitGate)
        self._entanglement_blocks = val
        
    @staticmethod
    def _parse_blocks(blocks) -> List:
        if not blocks:
            return []
        blocks = [blocks] if not isinstance(blocks, list) else blocks
        blocks = [QuantumCircuit._parse_gate_operation(block) \
                  if not isinstance(block, TemplateCircuitBlock) \
                  else block for block in blocks]
        return blocks
    
    @staticmethod
    def _validate_blocks(blocks, gate_feature=None, allow_template=True) -> None:
        for block in blocks:
            if isinstance(block, TemplateCircuitBlock):
                if not allow_template:
                    raise ValueError('template circuit block is not allowed'
                                     'in rotation layer')
                else:
                    continue
            elif isinstance(block, cirq.Gate):
                if gate_feature and \
                    not isinstance(block, gate_feature):
                    raise ValueError('Gate operation {} should be a subclass of '
                                     '{}'.format(type(block), gate_feature))
            elif callable(block):
                n_param = ParameterisedCircuit._get_parameter_count(block)
                if not n_param:
                    raise ValueError('not a valid gate operation or'
                                     'circuit block: {}'.format(block))
                dummy_params = tuple(sp.symarray('x',n_param))
                ParameterisedCircuit._validate_blocks([block(*dummy_params)], gate_feature)
        
    @staticmethod
    def _get_parameter_count(gate_expr) -> int:
        '''Returns the number of parameters required for a gate operation expression
        a gate operation expression can be a cirq.Gate instance, a string that can be mapped
        to a cirq.Gate instance, a function that maps to cirq.Gate or a TemplateCircuitBlock instance
        '''
        if isinstance(gate_expr, cirq.Gate):
            if isinstance(gate_expr, (cirq.XXPowGate, cirq.YYPowGate, cirq.ZZPowGate)):
                return 1
            return 0
        elif isinstance(gate_expr, TemplateCircuitBlock):
            return gate_expr.num_params
        elif callable(gate_expr):
            return len(signature(gate_expr).parameters)
        else:
            raise ValueError('invalid gate expression {} of type {}'.format(gate_expr, type(gate_expr)))

    def get_wavefunction(self, vals:np.ndarray):
        simulator = cirq.Simulator()
        param_resolver = self.get_parameter_resolver(vals)
        return simulator.simulate(self, param_resolver=param_resolver).final_state   
     
    
    def new_layer(self, name:str=None, *args, **kwargs):
        layer = super().new_layer(name, auto_assemble=False, *args, **kwargs)
        layer._parameter_symbol = self.parameter_symbol
        return layer

    def assemble(self):
        self.clear()
        cq = cirq.Circuit()
        for _ in range(self.copies):
            # can't just multiply since parameters won't get incremented
            cq += self.get_rotation_layer()
            cq += self.get_entanglement_layer()
        if self.final_rotation_layer:
            cq += self.get_rotation_layer()
            
        self.assign(cq)
        
        if self.flatten_circuit:
            self.flatten()
    
    def add_readout(self, gate:Union[str, cirq.Gate]='XX',
                    readout_qubit:Optional['cirq.GridQubit']=None):
        parsed_gate = QuantumCircuit._parse_gate_operation(gate)
        readout_qubit = cirq.GridQubit(-1, -1) if readout_qubit is None else readout_qubit
        for qubit in self.qubits:
            final_gate = self.parameterise_gate(parsed_gate)
            self.apply_gate_operation(final_gate, (readout_qubit, qubit))
        if self.flatten_circuit:
            self.flatten()            
        self._readout_qubit = readout_qubit
    
    def readout_measurement(self):
        if self.readout_qubit is None:
            raise ValueError('no readout qubit defined in the circuit')
        return cirq.Z(self.readout_qubit)
        
    def merge(self, other):
        self._moments = merge_pqc([self, other])._moments
        self._parameters = sp.symarray(self.parameter_symbol, len(self.symbols))
        if self.flatten_circuit:
            self.flatten()
                        
    def get_interaction_graphs(self, block_index:int, num_block_qubits:int,
                               entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                               *args, **kwargs):
        strategy = entangle_strategy if entangle_strategy is not None else self.entangle_strategy
        if isinstance(strategy, str):
            return interaction_graph[strategy](self.n_qubit, num_block_qubits)
        elif callable(strategy):
            return strategy(self.n_qubit, num_block_qubits)
        elif isinstance(strategy, list):
            if all(isinstance(strat, str) for strat in strategy):
                return interaction_graph[strategy[block_index]](self.n_qubit, num_block_qubits)
            elif all(callable(strat) for strat in strategy):
                return strategy[block_index](self.n_qubit, num_block_qubits)
        else:
            raise ValueError('invalid entangle strategy: {}'.format(strategy))
        
    # Devices are generally calibrated to circuits that alternate single-qubit gates with two-qubit gates in each layer. Staying close to this paradigm will often improve performance of circuits.

    def parameterise_gate(self, gate:Union[cirq.Gate, Callable]):
        if isinstance(gate, cirq.Gate):
            if isinstance(gate, (cirq.XXPowGate, cirq.YYPowGate, cirq.ZZPowGate)):
                param = self.new_param()
                return gate**param
            return gate
        else:
            n_param = ParameterisedCircuit._get_parameter_count(gate)
            params = self.new_param(size=n_param)
            params = tuple(params) if isinstance(params, np.ndarray) else (params,)
            return gate(*params)
        
    def add_rotation_layer(self, rotation_blocks:Optional[Union[str, cirq.Gate, Callable,
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None):
        rotation_blocks = self._parse_blocks(rotation_blocks)
        self._validate_blocks(rotation_blocks, SingleQubitGate, allow_template=False)    
        for block in rotation_blocks:
            for qubit in range(self.n_qubit):
                gate = self.parameterise_gate(block)
                self.apply_gate_operation(gate, qubit)
                
    def add_entanglement_layer(self, entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                               List[str],List[cirq.Gate],List[Callable],
                                               List['TemplateCircuitBlock']]] =None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None):
        entanglement_blocks = self._parse_blocks(entanglement_blocks)
        self._validate_blocks(entanglement_blocks, TwoQubitGate, allow_template=True)
        for i, block in enumerate(entanglement_blocks):
            if isinstance(block, TemplateCircuitBlock):
                interaction_graphs = self.get_interaction_graphs(i, block.num_block_qubits,
                                                                entangle_strategy)
                for qubits in interaction_graphs:
                    block.build(self, qubits)
            else:
                interaction_graphs = self.get_interaction_graphs(i, 2, entangle_strategy)
                for qubits in interaction_graphs:
                    gate = self.parameterise_gate(block)
                    self.apply_gate_operation(gate, qubits)
    
    def get_rotation_layer(self):
        rotation_layer = self.new_layer('rotation_layer')
        for rotation_block in self.rotation_blocks:
            for qubit in range(self.n_qubit):
                gate = self.parameterise_gate(rotation_block)
                rotation_layer.apply_gate_operation(gate, qubit)
        return rotation_layer

    def get_entanglement_layer(self):
        entanglement_layer = self.new_layer('entanglement_layer')
        for i, entanglement_block in enumerate(self.entanglement_blocks):
            if isinstance(entanglement_block, TemplateCircuitBlock):
                interaction_graphs = self.get_interaction_graphs(i, entanglement_block.num_block_qubits)
                for qubits in interaction_graphs:
                    entanglement_block.build(entanglement_layer, qubits)
            else:
                interaction_graphs = self.get_interaction_graphs(i, 2)
                for qubits in interaction_graphs:
                    gate = self.parameterise_gate(entanglement_block)
                    entanglement_layer.apply_gate_operation(gate, qubits)
    
        return entanglement_layer


    def run_simulator(self, repetitions:int=1):
        pass
        #simulator = cirq.Simulator()
        #simulator.run(resolved_circuit, repetitions=repetitions)
        ## to get wave function:
        #simulator.simulate(resolved_circuit)
        #cirq.measure_state_vector   
        #output_state_vector = simulator.simulate(self, resolver).final_state
        #z0 = cirq.X(q0)
        #qubit_map = {q0: 0, q1: 1}
        #z0.expectation_from_wavefunction(output_state_vector, qubit_map).real


from functools import reduce
import numpy as np
import sympy as sp

from quple.components.interaction_graphs import cyclic

def self_product(x: np.ndarray) -> float:
    """
    Define a function map from R^n to R.

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, x)  
    return coeff
    
def modified_cosine_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cos_x = [sp.cos(x_i) for x_i in x]
    coeff = cos_x[0] if len(x) == 1 else reduce(lambda m, n: m * n, cos_x)  
    return coeff

def cosine_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cos_x = [sp.cos(sp.pi*(x_i+1)/2) for x_i in x]
    coeff = cos_x[0] if len(x) == 1 else reduce(lambda m, n: m * n, cos_x)  
    return coeff

def sine_cosine_alternate_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    mix_x = [sp.cos(sp.pi*x[i]) if i%2 else sp.sin(sp.pi*x[i]) for i in range(len(x))]
    coeff = cos_x[0] if len(x) == 1 else reduce(lambda m, n: m * n, mix_x)  
    return coeff

def modified_sine_cosine_alternate_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    mix_x = [sp.cos(x[i]) if i%2 else sp.sin(sp.pi*x[i]) for i in range(len(x))]
    coeff = cos_x[0] if len(x) == 1 else reduce(lambda m, n: m * n, mix_x)  
    return coeff

def distance_measure(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0] - x[1])/2
    else:
        pairs = cyclic(len(x))
        y = [x[a]-x[b] for a,b in pairs]
        coeff = reduce(lambda m, n: m * n, y)/2**len(pairs)  
    return coeff


def distance_measure(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0] - x[1])/2
    else:
        pairs = cyclic(len(x))
        y = [x[a]-x[b] for a,b in pairs]
        coeff = reduce(lambda m, n: m * n, y)/2**len(pairs)  
    return coeff

def one_norm_distance(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = np.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [np.abs(x[a]-x[b]) for a,b in pairs]
        coeff = reduce(lambda m, n: m + n, y)/(2*len(pairs))
    return coeff

def two_norm_distance(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = np.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [(x[a] - x[b])**2 for a,b in pairs]
        coeff = (reduce(lambda m, n: m + n, y)**(1/2))/np.sqrt(len(pairs))
    return coeff

def arithmetic_mean(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m + n, x)/len(x) 
    return coeff    
    
def second_moment(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    square_x = [x_i**2 for x_i in x]
    coeff = x[0] if len(x) == 1 else (reduce(lambda m, n: m + n, square_x)/len(x))**0.5*2-1
    return coeff   

def third_moment(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cube_x = [x_i**3 for x_i in x]
    coeff = x[0] if len(x) == 1 else (abs(reduce(lambda m, n: m + n, cube_x))/len(x))**(1/3.0)*2-1
    return coeff 
                                          
def exponential_squared_sum(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    square_x = [x_i**2 for x_i in x]
    coeff = x[0] if len(x) == 1 else (np.e**(reduce(lambda m, n: m + n, square_x)/len(x))*2)/np.e-1
    return coeff         

def exponential_cube_sum(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    cube_x = [x_i**3 for x_i in x]
    coeff = x[0] if len(x) == 1 else (np.e**(reduce(lambda m, n: m + n, cube_x)/len(x))*2)/np.e-1
    return coeff     


from typing import Optional, Sequence, List, Dict, Union
from abc import ABC
import logging

import numpy as np
import six

import tensorflow as tf
import tensorflow_quantum as tfq

import cirq
import quple
from quple import ParameterisedCircuit

from pdb import set_trace


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)

class DefaultDict(dict):
    def __missing__(self, key):
        return key

class VQC(tf.keras.Sequential):
    def __init__(self, encoding_circuit:'cirq.Circuit', variational_circuit:'cirq.Circuit', 
                 optimizer:Optional[Union[str,tf.keras.optimizers.Optimizer]]='adam', 
                 differentiator:Optional[tfq.differentiators.Differentiator]=None,
                 loss='binary_crossentropy', 
                 activation='sigmoid',
                 metrics=['binary_accuracy'],
                 random_state:int=None,
                 readout=None,
                 trainable_dense_layer:bool=True,
                 name:str=None, *arg, **args):
        super(VQC, self).__init__()
        self._attributes = DefaultDict({})
        if random_state:
            tf.random.set_seed(random_state)
        self._readout = readout
        self.encoding_circuit = encoding_circuit
        self.variational_circuit = variational_circuit
        circuit_qubits = quple.get_circuit_qubits(variational_circuit)
        self._attributes['qubits'] = circuit_qubits
        self._attributes['n_qubit'] = len(circuit_qubits)
        self.differentiator = differentiator        
        self.activation = activation        
        self.random_state = random_state
        self.trainable_dense_layer = trainable_dense_layer
        layers = self._get_vqc_layers()
        super(VQC, self).__init__(layers, name)
        self.compile(loss=loss, optimizer=optimizer, metrics=metrics)
        
    def compile(self, optimizer='adam', loss='binary_crossentropy', metrics=['binary_accuracy'], *args, **kwargs):
        super(VQC, self).compile(optimizer, loss, metrics, *args, **kwargs)
        self._attributes['optimizer'] = self.optimizer._name if hasattr(self.optimizer, '_name') \
                                        else self.optimizer.__name__
        self._attributes['metrics'] = [metric._name if hasattr(metric, '_name')\
                                       else metric.__name__ for metric in self.metrics]
        self._attributes['loss'] = [loss.name if hasattr(loss, 'name') \
                                    else loss.__name__ for loss in self.loss_functions]
        
    @property
    def attributes(self):
        return self._attributes

    @property
    def encoding_circuit(self):
        return self._encoding_circuit
    
    @encoding_circuit.setter
    def encoding_circuit(self, val):
        feature_dimension = len(quple.get_circuit_unflattened_symbols(val))
        if not feature_dimension:
            raise ValueError('Encoding circuit must be a parameterised circuit with '
                             'number of parameters matching the feature dimension')
        self._check_circuit_qubits()
        self._encoding_circuit = val
        self._attributes['feature_dimension'] = feature_dimension
        self._attributes['encoding_circuit'] = val.name if isinstance(val, quple.QuantumCircuit) else ''
        self._attributes['encoding_map'] = val.encoding_map.__name__ if isinstance(val, quple.EncodingCircuit) else ''
        logger.info('Registered encoding circuit with feature dimension: {}'.format(feature_dimension))
    
    @property
    def qubits(self):
        return self.attributes['qubits']
    
    @property
    def n_qubit(self) -> int:
        return self.attributes['n_qubit']
        
    @property
    def symbols(self) -> List[str]:
        return self.attributes['symbols']
    
    @property
    def num_parameters(self) -> int:
        return self.attributes['num_parameters']
    
    @property
    def feature_dimension(self) -> int:
        return self.attributes['feature_dimension']
    
    @property
    def variational_circuit(self):
        return self._variational_circuit
    
    @variational_circuit.setter
    def variational_circuit(self, val):
        from pdb import set_trace
        circuit_parameters = quple.get_circuit_symbols(val)
        num_parameters = len(circuit_parameters)
        if not num_parameters:
            raise ValueError('Variational circuit must be a parameterised circuit which'
                             ' the parameters are to be optimized by the optimizer.')
        self._check_circuit_qubits()
        self._variational_circuit = val
        # if readout is not provided, default to measure all qubits
        if not self._readout:
            self._readout = [cirq.Z(qubit) for qubit in quple.get_circuit_qubits(val)]
        self._attributes['circuit_parameters'] = circuit_parameters
        self._attributes['num_parameters'] = num_parameters
        self._attributes['variational_circuit'] = val.name if isinstance(val, quple.QuantumCircuit) else ''
        logger.info('Registered variational circuit with number of parameters: {}'.format(num_parameters))
        
    def _check_circuit_qubits(self):
        # make sure both encoding and variational circuits are initialized
        if self.__dict__.get('_encoding_circuit', None) and \
        self.__dict__.get('_variational_circuit', None):
            encoding_circuit_qubits = quple.get_circuit_qubits(self.encoding_circuit)
            variational_circuit_qubits = quple.get_circuit_qubits(self.variational_circuit)
            if set(encoding_circuit_qubits) != set(variational_circuit_qubits):
                raise ValueError('encoding circuit and variational circuit must'
                                 ' have the same qubit layout')
    
    @property
    def readout(self):
        return self._readout
    
    @property
    def differentiator(self):
        return self._differentiator
    
    @differentiator.setter
    def differentiator(self, val):
        if (val != None) and (not isinstance(val, tfq.differentiators.Differentiator)):
            raise ValueError('Only tensorflow quantum differentiator is allowed')
        self._attributes['differentiator'] = val.__name__ if val else ''
        self._differentiator = val
    
    @property
    def activation(self):
        return self._activation
    
    @activation.setter
    def activation(self, val):
        if isinstance(val, six.string_types):
            val = tf.keras.activations.get(val)
        self._attributes['activation'] = val.__name__ if val else ''
        self._activation = val
        
    def _get_vqc_layers(self):
        
        input_layer = tf.keras.layers.Input(shape=(), dtype=tf.string)
        pqc_layer = tfq.layers.PQC(self.variational_circuit,
                                   self.readout,
                                   differentiator=self.differentiator)
        if self.trainable_dense_layer:
            output_layer = tf.keras.layers.Dense(1, activation=self.activation)
        else:
            output_layer = tf.keras.layers.Dense(1, activation=self.activation, 
                            kernel_initializer=tf.keras.initializers.Constant(value=1/self.n_qubit)) 
            output_layer.trainable = False
    
        return [input_layer, pqc_layer, output_layer]
    
    def _reset_layers(self):
        self._layers = None
        for layer in self._get_vqc_layers():
            self.add(layer)
        
    def _check_data(self, x):
        if isinstance(x, np.ndarray):
            num_dim = x.ndim
            if x.ndim != 2:
                raise ValueError('Data in numpy array format must be of dimension 2')
            num_var = x.shape[1]
            if num_var != self.feature_dimension:
                raise ValueError('Data has feature dimension {} but the encoding'
                ' circuit has feature dimension {}'.format(num_var, self.feature_dimension))
    
    def convert_to_tensor(self, x:np.ndarray):
        self._check_data(x)
        logger.info('Converting circuits to tensors...')
        return tfq.convert_to_tensor(self.encoding_circuit.resolve_parameters(x))
    
    def run(self, x_train, y_train, x_val, y_val, 
            x_test, y_test, 
            batch_size:Optional[int]=None,
            epochs:int=1, callbacks=None,
            roc_curve=True):
        self._attributes['train_size'] = len(x_train)
        self._attributes['val_size'] = len(x_val)
        self._attributes['test_size'] = len(x_test)
        self._attributes['batch_size'] = batch_size
        self._attributes['epochs'] = epochs   
        self.fit(x_train, y_train, batch_size, epochs,
                 validation_data=(x_val, y_val),
                 callbacks=callbacks)
        if isinstance(x_test, np.ndarray):
            x_test = self.convert_to_tensor(x_test)        
        self.evaluate(x_test, y_test, callbacks=callbacks) 
        if roc_curve:
            self.roc_curve(x_test, y_test, callbacks=callbacks)
        for callback in callbacks:
            if isinstance(callback, quple.VQCLogger):
                callback.reset_logger()
    def fit(self, x, y,
            batch_size:Optional[int]=None,
            epochs:int=1, 
            validation_data=None,
            *args, **kwargs):
        self._attributes['train_size'] = len(x)
        if isinstance(x, np.ndarray):
            x = self.convert_to_tensor(x)
        if validation_data and isinstance(validation_data, tuple):
            self._attributes['val_size'] = len(validation_data[0])
            if isinstance(validation_data[0], np.ndarray):
                x_val = self.convert_to_tensor(validation_data[0])
                validation_data = (x_val, validation_data[1])
        self._attributes['batch_size'] = batch_size
        self._attributes['epochs'] = epochs
        return super(VQC, self).fit(x, y,
                             batch_size=batch_size,
                             epochs=epochs,
                             validation_data=validation_data,
                             *args, **kwargs)
    
    def evaluate(self, x, y, *args, **kwargs):
        self._attributes['test_size'] = len(x)
        if isinstance(x, np.ndarray):
            x = self.convert_to_tensor(x)
        return super(VQC, self).evaluate(x, y, *args, **kwargs)
    
    def predict(self, x, *args, **kwargs):
        self._attributes['predict_size'] = len(x)
        if isinstance(x, np.ndarray):
            x = self.convert_to_tensor(x)
        return super(VQC, self).predict(x, *args, **kwargs)    
    
    def roc_curve(self, x, y, callbacks=None):
        y_true = y
        y_score = self.predict(x)
        from sklearn.metrics import roc_curve, auc
        fpr, tpr, thresholds = roc_curve(y_true, y_score)
        roc_auc = auc(fpr, tpr)
        for callback in callbacks:
            if isinstance(callback, quple.VQCLogger):
                callback.log_roc_curve(fpr, tpr, roc_auc)
        return fpr, tpr, thresholds, roc_auc
        